/*-------------------------------------------------------------------------*\
|                                                                           |
|   FILE:       ResizeResultManager.h										|
|   PURPOSE:    Declarations and definitions of classes to add a resize		|
|				behavior to functions that return array results.			|
|   DATE:       7 February 2013												|
|   VERSION:    7.0.7														|
|   COPYRIGHT:  (c) Planatech Solutions Ltd, 2013							|
|                                                                           |
|   This software is the property of Planatech Solutions Ltd.  It is        |
|   licensed for the use of the purchaser only, and no license is granted   |
|   for the further distribution of this source code whether for sale or    |
|   otherwise, whether in its original form or amended. Any binary files    |
|   generated from this source code can be distributed for sale to third    |
|   parties only with the express written permission of Planatech Solutions |
|   Ltd. Planatech Solutions Ltd and their agents accept no liability for   |
|   losses incurred of any kind through the use of this software.           |
|                                                                           |
|   UPDATE:     7 April 2016												|
|				Fixes to prevent crash under Excel 2010						|
|				23 August 2016												|
|				Support for omitting logging								|
\*-------------------------------------------------------------------------*/

/*
	The class offers a behavior such that array formulae are resized after 
	the Calculate event if the calling range is too small to hold the 
	array result.

	To use this class, declare a single instance at global scope, e.g:

	#include "ResizeResultManager.h"
	CResizeResultManager theResizeResultInstance;

	In each function that needs the behavior, make sure that the behavior 
	is optional, either (i) by including an extra boolean argument, typically 
	named resize or (ii) by having two versions of the function one which 
	resizes and another wihch does not. 
	
	This is important, because often the behavior will not be wanted, for 
	example when an array function is called within a function that produces 
	a scalar result, e.g. "=SUM(MyArrayFn(...))".

	To request the behavior, add a call to AddToQueueIfRequired() at the end 
	of the add-in function, e.g.:

	CXlOper* MyArrayFn_Impl(..., BOOL resize)
	{
		// Assign xloResult to an array
		xloResult.AllocArray(rows, cols);
		...
		
		if (resize)
		{
			theResizeResultInstance.AddToQueueIfRequired(xloResult);
		}
		return xloResult.Ret();
	}

	AddToQueueIfRequired() will compare the size of the caller to the size
	of the array in xloResult and will add it to a queue if the caller 
	is not large enough to hold the result. 

	After the next Calculate event, ProcessQueue() will be called and each 
	item in the queue will cause a resize behavior if this is possible.
*/

#ifndef __RESIZE_RESULTMANAGER_H__
#define __RESIZE_RESULTMANAGER_H__

#include <XlFuncs.h>
#include <XlpShared.h>

#ifndef __CPPLOG_H__
#include <cpplog_off.h>
#endif

class CResizeResultArgs
{
public:
	CResizeResultArgs(const CXlOper& xloCaller, 
		size_t nCallerRows, size_t nCallerCols, 
		size_t nRequiredRows, size_t nRequiredCols)
		: m_xloTopLeft(xloCaller.Resize(1, 1)), 
		  m_nCurrentRows(nCallerRows), m_nCurrentCols(nCallerCols),
		  m_nRequiredRows(nRequiredRows), m_nRequiredCols(nRequiredCols)
	{
		m_nNewRows = max(m_nRequiredRows, m_nCurrentRows); 
		m_nNewCols = max(m_nRequiredCols, m_nCurrentCols);
	}
public:
	CXlOper m_xloTopLeft;
	size_t m_nCurrentRows, m_nCurrentCols;
	size_t m_nRequiredRows, m_nRequiredCols;
	size_t m_nNewRows, m_nNewCols;
	CString m_strFormula;
};

class CResizeResultManager : public CXlLockable
{
private: 
	typedef std::vector<CResizeResultArgs> queue_t;
public:
	CResizeResultManager() 
	{
		m_calculateObserver.SetInstance(this);
	}
	void ProcessQueue()
	{
		LOG_TRACE(g_logger, "CResizeResultManager::ProcessQueue()");
		queue_t queue;
		MoveQueue(queue);
		for (size_t i = 0; i < queue.size(); i++)
		{
			Resize(queue[i]);
		}
	}
	void MoveQueue(queue_t& queue)	// Thread-safe
	{
		CXlLock lock(*this, true);
		queue = m_queue;
		m_queue.clear();
	}
	bool AddToQueueIfRequired(const CXlOper& xloResult)
	{
		LOG_TRACE(g_logger, "AddToQueueIfRequired(" << log(xloResult) << "): starts");
		CXlOper xloCaller;
		size_t nRequiredRows, nRequiredCols;
		size_t nCallerRows, nCallerCols;

		if (!xloResult.IsArray())
			return false;
		xloResult.GetDims(nRequiredRows, nRequiredCols);

		if (xloCaller.GetCaller() != 0 || !xloCaller.IsRef()) 
			return false;
		xloCaller.GetDims(nCallerRows, nCallerCols);

		if (nCallerRows < nRequiredRows || nCallerCols < nRequiredCols)
		{
			AddToQueue(xloCaller, nCallerRows, nCallerCols, nRequiredRows, nRequiredCols);
			return true;
		}
		else
		{
			LOG_TRACE(g_logger, "AddToQueueIfRequired(" << log(xloResult) << "): no action required");
		}
		return false;
	}
	void AddToQueue(const CXlOper& xloCaller, size_t nCallerRows, size_t nCallerCols, size_t nRequiredRows, size_t nRequiredCols) // Thread-safe
	{
		CXlLock lock(*this, true);
		m_queue.push_back(CResizeResultArgs(xloCaller, nCallerRows, nCallerCols, nRequiredRows, nRequiredCols));
		LOG_TRACE(g_logger, "AddToQueue(" << log(xloCaller) << ", " << nRequiredRows << ", " << nRequiredCols << ")");
	}
	void Resize(CResizeResultArgs& args)
	{
		LOG_TRACE(g_logger, "Resize starts: " << log(args.m_xloTopLeft) << " to [" << args.m_nNewRows << "," << args.m_nNewCols << "]");
		try
		{
			ResizeImpl(args);
#ifdef __CPPLOG_H__
			LOG_TRACE(g_logger, "Resized formula [" << args.m_strFormula << "] to [" << args.m_nNewRows << "," << args.m_nNewCols << "]");
#endif
		}
		catch(CXlRuntimeException& ex)
		{
#ifdef __CPPLOG_H__
			LOG_ERROR(g_logger, "Resize to [" << args.m_nNewRows << "," << args.m_nNewCols << "] failed : " << ex.what());
			if (args.m_strFormula.size() > 0)
				LOG_ERROR(g_logger, "Formula: [" << args.m_strFormula << "]");
#endif
		}
	}
	static CString log(const CXlOper& xlo)
	{
		if (xlo.IsRef())
		{
			return "CXlOper ref[" + xlo.GetRef().ToString() + "]";
		}
		else if (xlo.IsScalar())
		{
			return xlo.ToString();
		}
		else if (xlo.IsArray())
		{
			CString strBuf;
			size_t nRows, nCols;
			xlo.GetArraySize(nRows, nCols);
			strBuf.Format("CXlOper array[%lu rows x %lu cols]", nRows, nCols);
			return strBuf;
		}
		else
		{
			return "[CXlOper]";
		}
	}
protected:
	void ResizeImpl(CResizeResultArgs& args)
	{
		// If change is not required, quit
		if ((args.m_nCurrentRows >= args.m_nNewRows)
		 && (args.m_nCurrentCols >= args.m_nNewCols))
		{
			return;
		}

		int rc;
		bool bFormula;
		size_t r, c;
		
		// Get formula from top left cell
		if (!args.m_xloTopLeft.IsRef())
			throw CXlRuntimeException("Expected a reference");
		if ((rc = CXlMacros::GetCellIsFormula(bFormula, args.m_xloTopLeft)) != 0)
			throw CXlApiException("GetCellIsFormula", rc);
		if (!bFormula)
			throw CXlRuntimeException("Expected a reference to a cell containing a formula");
		if ((rc = CXlMacros::GetFormula(args.m_strFormula, args.m_xloTopLeft)) != 0)
			throw CXlApiException("GetFormula", rc);

		// Check that new cells in range are empty
		for (r = args.m_nCurrentRows; r < args.m_nNewRows; r++)
			for (c = 0; c < args.m_nNewCols; c++)
				CheckCellIsEmpty(args.m_xloTopLeft, r, c);
		for (c = args.m_nCurrentCols; c < args.m_nNewCols; c++)
			for (r = 0; r < args.m_nCurrentRows; r++)
				CheckCellIsEmpty(args.m_xloTopLeft, r, c);

		// Apply formula to new range
		CXlOper xloNewRange = args.m_xloTopLeft.Resize((RW12)args.m_nNewRows, (COL12)args.m_nNewCols);
		if ((rc = CXlMacros::FormulaArray(args.m_strFormula, xloNewRange)) != 0)
		{
			throw CXlApiException("FormulaArray", rc);
		}
	}
	void CheckCellIsEmpty(const CXlOper& xloTopLeft, size_t r, size_t c)
	{
		CString strFormula;
		BOOL bIsArray;
		CXlOper xloCell = xloTopLeft.Offset((RW12)r, (COL12)c);
		if (xloCell.GetFormula(strFormula, bIsArray)
		 && !strFormula.IsEmpty())
		{
			CString strMessage;
			strMessage.Format("cell [%ld, %ld] is not empty", r, c);
			throw CXlRuntimeException(strMessage);
		}
	}
private:
	queue_t m_queue;
private:
	// Events
	class CalculateObserver : public CXlCalculateEventStaticObserver
	{
	public:
		CalculateObserver() : m_instance(0) { }
		virtual void Update(CXlWorksheetEventArgs* e)
		{
			if (m_instance)
				m_instance->ProcessQueue();
		}
		void SetInstance(CResizeResultManager* instance) { m_instance = instance; }
	private:
		CResizeResultManager* m_instance;
	};
	CalculateObserver m_calculateObserver;
};

#endif // __RESIZE_RESULTMANAGER_H__

